/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#if !defined(_SOL_ARRAY)
#define _SOL_ARRAY
#define REVERSE true

class dict;

class sol_pair {
 public:
 text key;
 dyn  *value;

 sol_pair() { key = text(); value = new dyn(); }

 sol_pair(sol_pair *s) {
   key   = s->key;
   value = s->value;
 }

 sol_pair(int a, dyn b)  {
  key = text(a);
  value = new dyn(b);
 }

 sol_pair(const char *a, dyn b)  {
  key = text(a);
  value = new dyn(b);
 }

 sol_pair(text a, dyn b)  {
  key = a;
  value = new dyn(b);
 }

 text toText();
 char *toStr();

};


/**
  Associative, Php compatible dynamic array.
  Base class also of dict.
  */

class array {
 std::vector<text, std::allocator<text> > keyList;
 std::vector<dyn, std::allocator<dyn> > data;
 int freeKey;       // first free numeric key
 static dyn nullelement;	
	
 public:

 int indexer;

 /**
   Return the number of elements
   */
 inline int size()   { return (int) data.size(); }

 /**
   Return true if the array holds no any element, false otherwise
   */
 inline bool empty() { return data.empty(); }

 void clear();
 void set(const array &);
 void set(_intArr);
 void set(_arr32);
 void set(_array64);
 void set(array *);
 void set(dyn);
 void set(char *[]);

 array();

 /**
   Build an associative array from an array of strings of chars.
   Indexes are generated from 0 to n.
   */
 array(char *a[]);

 /**
   Build an array with a single element.
   */
 array(dyn);
 array(const array &);
 array(array *);

 /**
   Build an associative array from a typed indiced array.
   */
 
 array(_intArr);
 array(_arr32);
 array(_array64);

 ~array();


 void operator=(const array &);
 void operator=(array *);
 void operator=(dyn);
 void operator=(_intArr &);
 void operator=(_arr32 &);
 void operator=(_array64 &);



 void add(array *);
 void add(const array &);

 dyn get(const text &);
 dyn get(const char *);
 dyn get(int);

 inline array operator+(array x) { return operator+(&x); }
 array operator+(array *);
 array operator+(dyn);
 inline array operator-(array x) { return operator-(&x); }
 array operator-(array *);
 array operator-(dyn);
 inline array operator&(array x) { return operator&(&x); }
 array operator&(array *);
 inline array operator|(array x) { return operator|(&x); }
 array operator|(array *);
 inline array operator^(array x) { return operator^(&x); }
 array operator^(array *);

 inline array &operator+=(array x) { return operator+=(&x); }
 array &operator+=(array *);
 array &operator+=(dyn);
 void operator-=(array x) { operator-=(&x); }
 void operator-=(array *);
 void operator-=(dyn);
 inline bool operator==(array x) { return operator==(&x); }
 bool operator==(array *);
 bool operator==(bool);
 inline bool operator!=(array x) { return operator!=(&x); }
 bool operator!=(array *);
 bool operator!=(bool);
 inline bool operator<(array x) { return operator<(&x); }
 bool operator<(array *);
 inline bool operator>(array x) { return operator>(&x); }
 bool operator>(array *);
 inline bool operator>=(array a) { return !operator<(&a); }
 inline bool operator>=(array *a) { return !operator<(a); }
 inline bool operator<=(array b)  { return !operator>(&b); }
 inline bool operator<=(array *b) { return !operator>(b); }
 inline void operator&=(array x) { operator&=(&x); }
 void operator&=(array *);
 inline void operator|=(array x) { operator|=(&x); }
 void operator|=(array *);
 inline void operator^=(array x) { operator^=(&x); }
 void operator^=(array *);
 bool operator!();
 array &operator,(dyn);

 //===== util
 bool in(const dyn &);
 bool in(dyn *);
 signed int hasStrKey(const text &);
 signed int findKey(const text &);
 signed int findKey(const char *);
 signed int findKey(int);
 signed int findValue(const dyn &);
 signed int findValue(dyn *);
 inline int find(dyn x)  { return findValue(x); }
 inline int find(dyn *x) { return findValue(x); }
 signed int findLastValue(const dyn &);
 signed int findLastValue(dyn *);
 inline int findLast(dyn x)  { return findLastValue(x); }
 inline int findLast(dyn *x) { return findLastValue(x); }
 void display();
 void display(int);
 void pushKey(const text &);

 //===== iterator

 /**
   Restart the iterator (at first element).
   */
 void reset();
 bool outOfBound();
 bool bound();
 bool inBound();
 /**
   Return the current element, and move index to next one.
   */
 dyn inc();

 /**
   Return the current element, and move index to previous one.
   */
 dyn dec();

 /**
   Return the currently indiced element.
   */
 dyn current();

 /**
   Return the key of the currently pointed out element.
   */
 text key();

 /**
   Return the key of the element at position in argument.
   */
 text key(int);

 /**
   Return the value of the currently pointed out element.
   */
 dyn value();


 /**
   Return the value of the element at position in argument.
   */
 dyn value(int);

 /**
   Set iterator at first element and return it.
   */
 dyn begin();

 /**
   Set iterator at last element and return it.
   */
 dyn end();

 /**
   Return the current indice of the iterator.
   */
 int index();

 //===== stack

 /**
   Push a dyn on top of the array, it becomes the last element
   and the size increases from one.
   */
 void push(const dyn &);
 void push(dyn *);
 void push(sol_pair *);
 void push(const char *);
 void push(const text &, dyn);
 void pushAdd(const text &, dyn);
 void pushAdd(int, dyn);
 void pushUnic(const text &, dyn);

 /**
   Return the last element, and remove it from the array.
   If empty, return false.
   */
 dyn  pop();

 /**
   Return the first element and remove it.
   All remaining element are shift for one position. Indices are
   renumbered if integers.
   */
 dyn  shift();

 /**
   Insert an element at the first position.
   */
 void unshift(dyn);

 /**
   Insert an element at given position.
   The first parameter is the postion, the second one the element.
   */

 void insert(int, dyn);

 //===== ranges
 void erase(int, int);
 void erase(int);
 array slice(int, int, int);
 void splice(int, int, array, int);
 void splice(int, int, array *, int);
 void splice(int, int, int);
 void splice(int, int);
 void splice(int);
 void splice(int, int, dyn, int);
 void splice(int, int, bool, int);

 //===== transform

 /**
   Renumber the key if they are numbers. Textual keys are ignored.
   */
 void pack();

 void swapData(int, int);
 void swapElem(int, int);
 void keysSort(int, int);
 void dynSort(int, int);
 void dynRSort(int, int); 

 /**
   Reverse the content of the array
   */
 array reverse();

 /**
   Sort the values of the array in ascending order.
   */
 void sort(int x = 0);
 void rSort(int x = 0);

 /**
   Sort the keys of the array, preserving the associations with values.
   Ascending order.
   */
 void kSort(int x = 0);

 /**
   Remove doubloons in values.
   */
 array unique();

 /**
   Return the values as an array. Keys are generated, from 0 to n.
   */
 array values();

 /**
   Return the keys as an array. Integer keys are generated.
   */
 array keys();

 //===== operations
 void clone(array *);   // copy to another array

  /**
   Return the minimal value among elements.
   */
 dyn min();

 /**
   Return items selected by the function.
   */
 array map(bool (*)(dyn));

 /**
   Return the greatest values among elements.
   */
 dyn max();

 /**
   Calculate the sum of elements and return it.
   */
 double sum();

 /**
   Concatenate all the element (providing they are texts)
   and use the text given as argument as separator.
   */
 text join(const char *);

 int nomatch();

 //===== I/O

 /**
   Load a file into the array. The argument is the filename.
   Each line of the file become an element. The end of line code
   (10, 13 or the twice) are stored also.
   */
 void load(const char *);

 /**
   Save the whole array into the file, the name of which is given as
 argument. An end of line must be added to each element if necessary.
  */
 bool store(const char *);


 array *to(int);
 array *to(const char *);
 array *to(text);
 dyn &operator[](int);
 dyn &operator[](const char *);
 dyn &operator[](text);
 dyn &at(int);
 dyn &at(text);
 dyn &at(const char *);
 dyn &at(dyn);
 
    void setAt(int, dyn *);
    void setAt(int, dyn);
    void setAt(text, dyn *);
    void setAt(text, dyn);
    void unshift(text, dyn);

    char *toString(char *s = ", ");
    text toText(char *s = ", ");
	void boundError(const text &);

	friend class dict;
};


class dict : public array {
 public:

 /**
   Build a dict with a single element.
   */

    dict();
    dict(dyn x) : array(x) {};
    dict(const array &x) : array(x) {};
    dict(array *x) : array(x) {};
    dict(const dict &);
    dict(dict *);

    void set(const dict &);
    void set(dict *);
    void setAt(const text &, dyn);
    void setAt(const text &, dyn *);

    dyn &operator[](const char *);
    dyn &operator[](const text &);

    void erase(const char *);
    inline void erase(const text t) { erase(t.data); }
 
    void insert(int, dyn);
    void unshift(text, dyn);
    void kSort(int x = 0);
 
};


inline array operator+(array *a, const array &b) { return a->operator+(b); }


std::ostream& operator<< (std::ostream& out, array *);
std::ostream& operator<< (std::ostream& out, array);

#endif

