/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#ifndef _RWSTD_NO_NAMESPACE
  using namespace std;
#endif

#ifndef _SOL_DICT
#define _SOL_DICT



class sol_pair
{
 public:
 string key;
 dyn *value;

 sol_pair() { key = "", value = new dyn(); }
 sol_pair(sol_pair *s)
 {
   key   = s->key;
   value = s->value;
 }
 sol_pair(int a, dyn b)
 {
  char buf[32];
  sprintf(buf, "%d", a);
  key = buf;
  value = new dyn(b);
 }
 sol_pair(string a, dyn b)
 {
  key = a;
  value = new dyn(b);
 }

};


class dict
{
 vector<string> keys;
 vector<dyn> data;
 int freeKey;       // first free numeric key

 public:
 int index;
 inline int size()   { return (int) data.size(); }
 inline bool empty() { return data.empty(); }

 void clear();
 void set(dict);
 void set(dict *);
 void set(dyn);
// template<class ANY> void operator=(ANY a)   { set(a); }
 void operator=(dict a)  { set(a); }
 void operator=(dict *a) { set(a); }
 void operator=(dyn a)    { set(a); }
 template<class ANY> void assign(int idx, ANY a)
 {
   int k = findKey(idx);
   if(k != -1) data[k] = dyn(a);
   push(idx, dyn(a));
 }
template<class ANY> void assign(string idx, ANY a)
 {
   int k = findKey(idx);
   if(k != -1) data[k] = dyn(a);
   push(idx, dyn(a));
 }



 dict();
 dict(int, ... );
 template<class ANY> dict(ANY a) { set(a); }

 void add(dict);
 void add(dict *);
 void add(dyn);
 template<class ANY> void operator+=(ANY a) { add(a); }

 dyn get(const string &);
 dyn get(const char *);
 dyn get(int);

 dict operator+(dict);
 dict operator+(dict *);
 dict operator+(dyn);
 dict operator-(dict);
 dict operator-(dict *);
// dict operator-(dyn);
 dict operator&(dict);
 dict operator&(dict *);
 dict operator|(dict);
 dict operator|(dict *);
 void operator-=(dict);
 void operator-=(dict *);
 bool operator==(dict);
 bool operator==(dict *);
 bool operator!=(dict);
 bool operator!=(dict *);
 bool operator<(dict);
 bool operator<(dict *);
 bool operator>(dict);
 bool operator>(dict *);
 inline bool operator>=(dict a) { return !operator<(a); }
 inline bool operator>=(dict *a) { return !operator<(a); }
 inline bool operator<=(dict b)  { return !operator>(b); }
 inline bool operator<=(dict *b) { return !operator>(b); }
 void operator&=(dict);
 void operator&=(dict *);
 void operator|=(dict);
 void operator|=(dict *);

 inline dyn  operator[](int a) { return get(a); }
 inline dyn  operator[](const string &a) { return get(a); }
 inline dyn  operator[](const char *a) { return get(a); }
 dyn & dict  operator[](text t);


 //===== util
 bool in(dyn *d);
 signed int hasStrKey(const string &);
 signed int findKey(const string &);
 signed int findKey(int);
 signed int findValue(dyn);
 signed int findValue(dyn *);
 void display();
 string toString();
 void pushKey(const string &);

 void renumber();

 //===== iterator
 void reset();
 dyn inc();
 dyn dec();
 dyn current();
 dyn key();
 dyn key(int);
 dyn value(int);
 dyn begin();
 dyn end();

 //===== stack
 void push(dyn);
 void push(dyn *);
 void push(sol_pair *);
 void push(const string &, dyn);
 void push(int, dyn);
 void pushUnic(string &, dyn);
 dyn  pop();
 dyn  shift();
 void unshift(dyn);

 //===== ranges
 void erase(int, int);
 void erase(int);

 void splice(int, int, dict);
 void splice(int, int, dict *);
 void splice(int, int);
 void splice(int);
 void splice(int, int, dyn);

 //===== transform

 void swap(int, int);
 void kSort(int, int);
 void dynSort(int, int);
 void sort();
 void ksort();
 void unique();
 array values();

 //===== operations
 void clone(dict *);   // copy to another dict
 dyn min();
 dyn max();
 double sum();
 string join(const char *);

 int nomatch();
};


ostream& operator<< (ostream& out, dict *);
ostream& operator<< (ostream& out, dict);


#endif

