/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <ctype.h>
#include <sys/stat.h>
#include <time.h>
#include <math.h>
#include <dirent.h>

#ifdef UNIX
#include <unistd.h>
#else
#include <io.h>
#endif

#include <iostream>
#include <vector>

#include "text.hpp"
#include "intarr.hpp"
#include "textarr.hpp"
#include "array64.hpp"
#include "dyn.hpp"
#include "array.hpp"
#include "fun.hpp"
#include "array.hpp"

//===== array

int count(array a) { return a.size(); }

array  _sol_range(int a, int b)
{
  array x = array();

  if(a > b)
  {
    while(a > b)
    {
     x.push(dyn(a));
     a--;
    }
  }
  else
  {
    while(a <= b)
    {
     x.push(dyn(a));
     a++;
    }
  }
  return x;
}

array array_merge(array a, array b) { return a + b; }
array array_diff(array a, array b) { return a - b; }
array array_intersect(array a, array b) { return a & b; }
array array_unique(const array &a) { array b = a; b.unique(); return b; }

bool in_array(int i, array a)    { return a.in(new dyn(i)); }
bool in_array(text s, array a) { return a.in(new dyn(s)); }
bool in_array(const char s, array a) { return a.in(new dyn(s)); }
bool in_array(double d, array a) { return a.in(new dyn(d)); }
bool in_array(dyn d, array a)    { return a.in(d); }

// for control structure using either text or array
bool in_sequence(const char *x, text &t) { return t.in(x); }
bool in_sequence(const char *x, const char *t) { return text(t).in(x); }
//bool in_sequence(dyn d, array a) { return a.in(d); }
//bool in_sequence(text d, array a) { return a.in(dyn(d)); }



// one argument
/*
template<class R> void array_map(R(*fun)(dyn &), array &a)
{
  a.begin();
  for(int i = 0; i < a.size(); i++)
   {
     fun(a[a.key()]);
     a.inc();
   }
}

template<class R> void array_map(R(*fun)(dyn *), array &a)
{
  a.begin();
  for(int i = 0; i < a.size(); i++)
   {
     fun(a[a.key()]);
     a.inc();
   }
}
*/

void array_map(void(*fun)(dyn &), array &a)
{
  a.begin();
  for(int i = 0; i < a.size(); i++)
   {
     fun(a[a.key()]);
     a.inc();
   }
}

void array_map(void(*fun)(dyn *), array &a)
{
  a.begin();
  for(int i = 0; i < a.size(); i++)
   {
     fun(&a[a.key()]);
     a.inc();
   }
}


void array_map(void (*fun)(dyn &, dyn &), array &a, array &b)
{
  int i = 0;
  while(i < a.size() && i < b.size())
  {
    if(i == 0) { a.begin(); b.begin();  }
    else      { a.inc(); b.inc(); }
    fun(a[a.key()], b[b.key()]);
    i++;
  }
}

void array_map(void (*fun)(dyn &a, dyn &b, dyn &c), array &a, array &b, array &c)
{
  int i = 0;
  while(i < a.size() && i < b.size() && i < c.size())
  {
    if(i == 0) { a.begin(); b.begin(); c.begin();  }
    else      { a.inc(); b.inc(); c.inc(); }
    fun(a[a.key()], b[b.key()], c[c.key()]);
    i++;
  }
}


//===== iterators

dyn reset(array &x) { return x.begin(); }
dyn next(array &x)  { return x.inc(); }
dyn prev(array &x)  { return x.dec(); }
dyn end(array &x)   { return x.end(); }
text key(array &x)  { return x.key(); }
dyn current(array &x) { return x.value(); }

//===== functions on arrays

text implode(array a, const char *s) { return a.join(s); }

/*
array explode(const char *s, const char *sep)
{
  int pos;
  array x;
  const char *nstr;
  int len;
  int off;
  int slen;
  int olen;

  olen = strlen(s);
  slen = strlen(sep);
  off = 0;

loop:
  //cout << "search in " << s << " at " <<  off << endl;
  pos = strpos(s, sep, off);
  //cout << "pos=" << pos << endl;
  if(pos != -1)
  {
    len = pos - off;
    nstr = substr(s, off, len);
    x.push(nstr);
    //cout << "added " << nstr << " pos=" << pos << " len=" << len << " slen=" << slen << endl;
    pos += slen;
    off = pos;
    if(off < olen) goto loop;
  }

  if(off < olen)
  {
    x.push(s + off);
    //cout << "added last " << (s + off) << endl;
  }
  return x;
}
*/

array explode(const char *s, const char *sep)
{
  array x;
  const char *nstr;
  int olen = strlen(s);
  char *temp = (char *) malloc(olen + 1);
  strcpy(temp, s);
  nstr = strtok(temp, sep);
  
loop:  

  if(nstr != NULL)
  {
    //printf("next %s\n", nstr);
    x.push(nstr);
    nstr = strtok(NULL, sep);
    goto loop;
  }

  return x;
}


//===== files


array file(text name) { return file(name.data); }
array file(const char *name)
{
 FILE *f;
 char BUFFER[65536];
 char *ptr;
 array a = array();

 f = fopen(name, "r");
 if(f == NULL) return array();
 while(1)
 {
  ptr = fgets(BUFFER, 65536, f);
  if(ptr == NULL) break;
  a.push(dyn(ptr));
 }
 fclose(f);
 return a;
}

void store(array a, text name) { store(a, name.data); }
void store(array a, const char *name)
{
 FILE *f;

 f = fopen(name, "w");
 if(f == NULL) return;
 dyn x;
 a.begin();
 do
 {
   x.set(a.current());
   fputs(x.toStr(), f);
 }
 while(!dyn::isNil(a.inc()));

 fclose(f);
}

char *extractProperty(const char *t, const char *prop)
{
  const char *x;
  int pos = text(t).findLex(prop);
  if(pos == -1) return "";
  pos += strlen(prop);

  x = t + pos;
  while(*x == ' ' || *x == '=' || *x == '\'' || *x == '\"') x++;
  pos = 0;
  while(x[pos] != '\'' && x[pos] != '\"' && x[pos] != '>') pos++;
  return substr(x, 0, pos);
}

array get_meta_tags(const char *fname)
{
  array a = file(fname);
  array b = array();
  int pos, eot;
  bool inside = false;
  text t, metas;

  a.begin();
  do
  {
   t = a.current().toText();
   if(!inside)
   {
     pos = t.findLex("<head>");
     if(pos != -1)
     {
       t = t.slice(pos + 6, t.length(), 0);
       inside = true;
     }
   }
   if(inside)
   {
     pos = t.findLex("</head>");
     if(pos != -1)
     {
       b.push(t.slice(0, pos, 0));
       break;
     }
     b.push(t.trim());
   }
  } while(!dyn::isNil(a.inc()));

  t = b.join("");

// now search metas inside the text

  a.clear();

  pos = 0;
  while(pos != -1)
  {
    pos = t.findLex("<meta", pos);
    if(pos != -1)
    {
      eot = t.find(">", pos + 5);
      if(eot != -1)
      {
        a.push(dyn(t.slice(pos, eot, 1)));
      }
      pos = eot;
    }
  }

 b.clear();
 for(int i = 0; i < a.size(); i++)
 {
   t = a[i].toText();
   metas = extractProperty(t.data, "name");
   if(metas.len > 0)
    b[metas] = extractProperty(t.data, "content");
 }

 return b;
}


array array_reverse(array x, bool preserve)
{
  array y;
  int len = x.size() - 1;
  int middle = len / 2;
  y.set(x);
  for(int i = 0; i < middle; i++)
  {
    if (preserve == true) y.swapElem(i, len - i);
                     else y.swapData(i, len - i);
  }
  return y;
}

// mut the content of a dir into an array

array _scandir(DIR *x)	
{
	array a = array();

	while(1)
	{	
		struct dirent *de = readdir(x);
		if(de == NULL) break;	
		char *str = de->d_name;	
		if(str != NULL)
		{	
			if(strlen(str) > 0)
			{	
				if( str[0] != '.')  
				{	
					a.push(dyn(str));
				}	
				continue;
			}	
		}	
		break;
	}
	return a;	
}	


char *str(array a)
{
	return a.toString();
}
